// Emacs style mode select   -*- C++ -*- 
//-----------------------------------------------------------------------------
//
// $Id:$
//
// Copyright (C) 1993-1996 by id Software, Inc.
//
// This source is available for distribution and/or modification
// only under the terms of the DOOM Source Code License as
// published by id Software. All rights reserved.
//
// The source is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// FITNESS FOR A PARTICULAR PURPOSE. See the DOOM Source Code License
// for more details.
//
// $Log:$
//
// DESCRIPTION:
//	Archiving: SaveGame I/O.
//
//-----------------------------------------------------------------------------

static const char
rcsid[] = "$Id: p_tick.c,v 1.4 1997/02/03 16:47:55 b1 Exp $";

#include "doomdef.h"

#include "i_system.h"
#include "z_zone.h"
#include "p_local.h"

// State.
#include "doomstat.h"
#include "r_state.h"

byte*		save_p;
extern byte *savebuffer;

// Pads save_p to a 4-byte boundary
//  so that the load/save works on SGI&Gecko.
//#define PADSAVEP()	save_p += (4 - ((int) save_p & 3)) & 3
#define PADSAVEP()		save_p += (4 - ((save_p - savebuffer) & 3)) & 3
//#define PADSAVEP()		save_p += 3



// Dull, but inevitable.
void P_PlayerConverter(player_t *plr, saveplayer_t *saveplr, boolean saving)
{
	if(saving)
	{
		// Some data (message, attacker, etc.) is skipped.
		saveplr->playerstate = plr->playerstate;
		saveplr->viewz = plr->plr->viewz;
		saveplr->viewheight = plr->viewheight;
		saveplr->deltaviewheight = plr->deltaviewheight;
		saveplr->bob = plr->bob;
		saveplr->health = plr->health;
		saveplr->armorpoints = plr->armorpoints;
		saveplr->armortype = plr->armortype;
		memcpy(saveplr->powers, plr->powers, sizeof(saveplr->powers));
		memcpy(saveplr->cards, plr->cards, sizeof(saveplr->cards));
		saveplr->backpack = plr->backpack;
		memcpy(saveplr->frags, plr->plr->frags, sizeof(saveplr->frags));
		saveplr->readyweapon = plr->readyweapon;
		saveplr->pendingweapon = plr->pendingweapon;
		memcpy(saveplr->weaponowned, plr->weaponowned, sizeof(saveplr->weaponowned));
		memcpy(saveplr->ammo, plr->ammo, sizeof(saveplr->ammo));
		memcpy(saveplr->maxammo, plr->maxammo, sizeof(saveplr->maxammo));
		saveplr->attackdown = plr->attackdown;
		saveplr->usedown = plr->usedown;
		saveplr->cheats = plr->cheats;
		saveplr->refire = plr->refire;
		saveplr->killcount = plr->killcount;
		saveplr->itemcount = plr->itemcount;
		saveplr->secretcount = plr->secretcount;
		saveplr->damagecount = plr->damagecount;
		saveplr->bonuscount = plr->bonuscount;
		saveplr->extralight = plr->plr->extralight;
		saveplr->fixedcolormap = plr->plr->fixedcolormap;
		saveplr->colormap = plr->colormap;
		memcpy(saveplr->psprites, plr->psprites, sizeof(saveplr->psprites));
		saveplr->didsecret = plr->didsecret;
	}
	else // loading
	{
		plr->playerstate = saveplr->playerstate;
		plr->plr->viewz = saveplr->viewz;
		plr->viewheight = saveplr->viewheight;
		plr->deltaviewheight = saveplr->deltaviewheight;
		plr->bob = saveplr->bob;
		plr->health = saveplr->health;
		plr->armorpoints = saveplr->armorpoints;
		plr->armortype = saveplr->armortype;
		memcpy(plr->powers, saveplr->powers, sizeof(plr->powers));
		memcpy(plr->cards, saveplr->cards, sizeof(plr->cards));
		plr->backpack = saveplr->backpack;
		memcpy(plr->plr->frags, saveplr->frags, sizeof(saveplr->frags));
		plr->readyweapon = saveplr->readyweapon;
		plr->pendingweapon = saveplr->pendingweapon;
		memcpy(plr->weaponowned, saveplr->weaponowned, sizeof(plr->weaponowned));
		memcpy(plr->ammo, saveplr->ammo, sizeof(plr->ammo));
		memcpy(plr->maxammo, saveplr->maxammo, sizeof(plr->maxammo));
		plr->attackdown = saveplr->attackdown;
		plr->usedown = saveplr->usedown;
		plr->cheats = saveplr->cheats;
		plr->refire = saveplr->refire;
		plr->killcount = saveplr->killcount;
		plr->itemcount = saveplr->itemcount;
		plr->secretcount = saveplr->secretcount;
		plr->damagecount = saveplr->damagecount;
		plr->bonuscount = saveplr->bonuscount;
		plr->plr->extralight = saveplr->extralight;
		plr->plr->fixedcolormap = saveplr->fixedcolormap;
		plr->colormap = saveplr->colormap;
		memcpy(plr->psprites, saveplr->psprites, sizeof(plr->psprites));
		plr->didsecret = saveplr->didsecret;
	}
}

// Dull, but inevitable.
void P_MobjConverter(mobj_t *mobj, savemobj_t *savemobj, boolean saving)
{
	if(saving)
	{
		memset(savemobj, 0, sizeof(*savemobj));

		savemobj->thinker = mobj->thinker;
		savemobj->x = mobj->x;
		savemobj->y = mobj->y;
		savemobj->z = mobj->z;
		savemobj->angle = mobj->angle;
		savemobj->sprite = mobj->sprite;
		savemobj->frame = mobj->frame;
		savemobj->floorz = mobj->floorz;
		savemobj->ceilingz = mobj->ceilingz;
		savemobj->radius = mobj->radius;
		savemobj->height = mobj->height;
		savemobj->momx = mobj->momx;
		savemobj->momy = mobj->momy;
		savemobj->momz = mobj->momz;
		savemobj->Validcount = mobj->Validcount;
		savemobj->type = mobj->type;
		savemobj->tics = mobj->tics;
		savemobj->state = mobj->state;
		savemobj->flags = mobj->flags;
		savemobj->health = mobj->health;
		savemobj->movedir = mobj->movedir;
		savemobj->movecount = mobj->movecount;
		savemobj->reactiontime = mobj->reactiontime;
		savemobj->threshold = mobj->threshold;
		savemobj->player = mobj->player;
		savemobj->lastlook = mobj->lastlook;
		savemobj->spawnpoint = mobj->spawnpoint;
	}
	else // loading
	{
		memset(mobj, 0, sizeof(*mobj));

		mobj->thinker = savemobj->thinker;
		mobj->x = savemobj->x;
		mobj->y = savemobj->y;
		mobj->z = savemobj->z;
		mobj->angle = savemobj->angle;
		mobj->sprite = savemobj->sprite;
		mobj->frame = savemobj->frame;
		mobj->floorz = savemobj->floorz;
		mobj->ceilingz = savemobj->ceilingz;
		mobj->radius = savemobj->radius;
		mobj->height = savemobj->height;
		mobj->momx = savemobj->momx;
		mobj->momy = savemobj->momy;
		mobj->momz = savemobj->momz;
		mobj->Validcount = savemobj->Validcount;
		mobj->type = savemobj->type;
		mobj->tics = savemobj->tics;
		mobj->state = savemobj->state;
		mobj->flags = savemobj->flags;
		mobj->health = savemobj->health;
		mobj->movedir = savemobj->movedir;
		mobj->movecount = savemobj->movecount;
		mobj->reactiontime = savemobj->reactiontime;
		mobj->threshold = savemobj->threshold;
		mobj->player = savemobj->player;
		mobj->lastlook = savemobj->lastlook;
		mobj->spawnpoint = savemobj->spawnpoint;
	}
}


//
// P_ArchivePlayers
//
void P_ArchivePlayers (void)
{
    int			i;
    int			j;
	saveplayer_t *dest;
	
    for(i=0; i<MAXPLAYERS; i++)
    {
		if (!players[i].plr->ingame)
			continue;
		
		PADSAVEP();
		
		dest = (saveplayer_t*) save_p;
		P_PlayerConverter(&players[i], dest, true);
		save_p += sizeof(saveplayer_t);
		for(j=0; j<NUMPSPRITES; j++)
		{
			if(dest->psprites[j].state)
			{
				dest->psprites[j].state 
					= (state_t*) (dest->psprites[j].state-states);
			}
		}
    }
}



//
// P_UnArchivePlayers
//
void P_UnArchivePlayers (void)
{
    int		i;
    int		j;
	
    for (i=0 ; i<MAXPLAYERS ; i++)
    {
		if (!players[i].plr->ingame)
			continue;
		
		PADSAVEP();
		
		//memcpy (&players[i],save_p, sizeof(player_t));

		P_PlayerConverter(&players[i], (saveplayer_t*) save_p, false);
		save_p += sizeof(saveplayer_t);
		
		// will be set when unarc thinker
		players[i].plr->mo = NULL;	
		players[i].message = NULL;
		players[i].attacker = NULL;
		
		for(j=0; j<NUMPSPRITES; j++)
		{
			if(players[i].psprites[j].state)
			{
				players[i].psprites[j].state 
					= &states[ (int) players[i].psprites[j].state ];
			}
		}
    }
}


//
// P_ArchiveWorld
//
void P_ArchiveWorld (void)
{
    int			i;
    int			j;
    sector_t*	sec;
    line_t*		li;
    side_t*		si;
    short*		put;
	
    put = (short*) save_p;
    
    // do sectors
    for(i=0, sec = sectors; i<numsectors; i++, sec++)
    {
		*put++ = sec->floorheight >> FRACBITS;
		*put++ = sec->ceilingheight >> FRACBITS;
		*put++ = sec->floorpic;
		*put++ = sec->ceilingpic;
		*put++ = sec->lightlevel;
		*put++ = sec->special;		// needed?
		*put++ = sec->tag;		// needed?
    }
    
    // do lines
    for(i=0, li = lines; i<numlines; i++, li++)
    {
		*put++ = li->flags;
		*put++ = li->special;
		*put++ = li->tag;
		for(j=0; j<2; j++)
		{
			if (li->sidenum[j] == -1)
				continue;
			
			si = &sides[li->sidenum[j]];
			
			*put++ = si->textureoffset >> FRACBITS;
			*put++ = si->rowoffset >> FRACBITS;
			*put++ = si->toptexture;
			*put++ = si->bottomtexture;
			*put++ = si->midtexture;	
		}
    }
    save_p = (byte*) put;
}



//
// P_UnArchiveWorld
//
void P_UnArchiveWorld (void)
{
    int			i;
    int			j;
    sector_t*	sec;
    line_t*		li;
    side_t*		si;
    short*		get;
	
    get = (short*) save_p;
    
    // do sectors
    for (i=0, sec = sectors ; i<numsectors ; i++,sec++)
    {
		sec->floorheight = *get++ << FRACBITS;
		sec->ceilingheight = *get++ << FRACBITS;
		sec->floorpic = *get++;
		sec->ceilingpic = *get++;
		sec->lightlevel = *get++;
		sec->special = *get++;		// needed?
		sec->tag = *get++;		// needed?
		sec->specialdata = 0;
		sec->soundtarget = 0;
    }
    
    // do lines
    for (i=0, li = lines ; i<numlines ; i++,li++)
    {
		li->flags = *get++;
		li->special = *get++;
		li->tag = *get++;
		for (j=0 ; j<2 ; j++)
		{
			if (li->sidenum[j] == -1)
				continue;
			si = &sides[li->sidenum[j]];
			si->textureoffset = *get++ << FRACBITS;
			si->rowoffset = *get++ << FRACBITS;
			si->toptexture = *get++;
			si->bottomtexture = *get++;
			si->midtexture = *get++;
		}
    }
    save_p = (byte *)get;	
}





//
// Thinkers
//
typedef enum
{
    tc_end,
    tc_mobj

} thinkerclass_t;



//
// P_ArchiveThinkers
//
void P_ArchiveThinkers (void)
{
    thinker_t*	th;
    savemobj_t*	smobj;
	
    // save off the current thinkers
    for(th = thinkercap.next ; th != &thinkercap ; th=th->next)
    {
		if(th->function == P_MobjThinker)
		{
			*save_p++ = tc_mobj;
			PADSAVEP();
			smobj = (savemobj_t*) save_p;
			//memcpy (mobj, th, sizeof(*mobj));
			P_MobjConverter( (mobj_t*) th, smobj, true);
			save_p += sizeof(*smobj);
			smobj->state = (state_t *)(smobj->state - states);
			
			if (smobj->player)
				smobj->player = (player_t *)((smobj->player-players) + 1);
			continue;
		}
		
		// I_Error ("P_ArchiveThinkers: Unknown thinker function");
    }
	
    // add a terminating marker
    *save_p++ = tc_end;	
}



//
// P_UnArchiveThinkers
//
void P_UnArchiveThinkers (void)
{
    byte		tclass;
    thinker_t*	currentthinker;
    thinker_t*	next;
    mobj_t*		mobj;
    
    // remove all the current thinkers
    currentthinker = thinkercap.next;
    while (currentthinker != &thinkercap)
    {
		next = currentthinker->next;
		
		if (currentthinker->function == P_MobjThinker)
			P_RemoveMobj ((mobj_t *)currentthinker);
		else
			Z_Free (currentthinker);
		
		currentthinker = next;
    }
    P_InitThinkers ();
	
    // read in saved thinkers
    while (1)
    {
		tclass = *save_p++;
		switch (tclass)
		{
		case tc_end:
			return; 	// end of list
			
		case tc_mobj:
			PADSAVEP();
			mobj = Z_Malloc (sizeof(*mobj), PU_LEVEL, NULL);
			//memcpy (mobj, save_p, sizeof(*mobj));
			P_MobjConverter(mobj, (savemobj_t*) save_p, false);
			save_p += sizeof(savemobj_t);
			mobj->state = &states[(int)mobj->state];
			mobj->target = NULL;
			if (mobj->player)
			{
				mobj->player = &players[(int)mobj->player-1];
				mobj->player->plr->mo = mobj;
			}
			P_SetThingPosition (mobj);
			mobj->info = &mobjinfo[mobj->type];
			mobj->floorz = mobj->subsector->sector->floorheight;
			mobj->ceilingz = mobj->subsector->sector->ceilingheight;
			mobj->thinker.function = P_MobjThinker;
			P_AddThinker (&mobj->thinker);
			break;
			
		default:
			I_Error ("Unknown tclass %i in savegame",tclass);
		}
    }
}


//
// P_ArchiveSpecials
//
enum
{
    tc_ceiling,
    tc_door,
    tc_floor,
    tc_plat,
    tc_flash,
    tc_strobe,
    tc_glow,
    tc_endspecials

} specials_e;	



//
// Things to handle:
//
// T_MoveCeiling, (ceiling_t: sector_t * swizzle), - active list
// T_VerticalDoor, (vldoor_t: sector_t * swizzle),
// T_MoveFloor, (floormove_t: sector_t * swizzle),
// T_LightFlash, (lightflash_t: sector_t * swizzle),
// T_StrobeFlash, (strobe_t: sector_t *),
// T_Glow, (glow_t: sector_t *),
// T_PlatRaise, (plat_t: sector_t *), - active list
//
void P_ArchiveSpecials (void)
{
    thinker_t*		th;
    ceiling_t*		ceiling;
    vldoor_t*		door;
    floormove_t*	floor;
    plat_t*			plat;
    lightflash_t*	flash;
    strobe_t*		strobe;
    glow_t*			glow;
    int				i;
	
    // save off the current thinkers
    for (th = thinkercap.next ; th != &thinkercap ; th=th->next)
    {
		if (th->function == NULL)
		{
			for (i = 0; i < MAXCEILINGS;i++)
				if (activeceilings[i] == (ceiling_t *)th)
					break;
				
				if (i<MAXCEILINGS)
				{
					*save_p++ = tc_ceiling;
					PADSAVEP();
					ceiling = (ceiling_t *)save_p;
					memcpy (ceiling, th, sizeof(*ceiling));
					save_p += sizeof(*ceiling);
					ceiling->sector = (sector_t *)(ceiling->sector - sectors);
				}
				continue;
		}
		
		if (th->function == T_MoveCeiling)
		{
			*save_p++ = tc_ceiling;
			PADSAVEP();
			ceiling = (ceiling_t *)save_p;
			memcpy (ceiling, th, sizeof(*ceiling));
			save_p += sizeof(*ceiling);
			ceiling->sector = (sector_t *)(ceiling->sector - sectors);
			continue;
		}
		
		if (th->function == T_VerticalDoor)
		{
			*save_p++ = tc_door;
			PADSAVEP();
			door = (vldoor_t *)save_p;
			memcpy (door, th, sizeof(*door));
			save_p += sizeof(*door);
			door->sector = (sector_t *)(door->sector - sectors);
			continue;
		}
		
		if (th->function == T_MoveFloor)
		{
			*save_p++ = tc_floor;
			PADSAVEP();
			floor = (floormove_t *)save_p;
			memcpy (floor, th, sizeof(*floor));
			save_p += sizeof(*floor);
			floor->sector = (sector_t *)(floor->sector - sectors);
			continue;
		}
		
		if (th->function == T_PlatRaise)
		{
			*save_p++ = tc_plat;
			PADSAVEP();
			plat = (plat_t *)save_p;
			memcpy (plat, th, sizeof(*plat));
			save_p += sizeof(*plat);
			plat->sector = (sector_t *)(plat->sector - sectors);
			continue;
		}
		
		if (th->function == T_LightFlash)
		{
			*save_p++ = tc_flash;
			PADSAVEP();
			flash = (lightflash_t *)save_p;
			memcpy (flash, th, sizeof(*flash));
			save_p += sizeof(*flash);
			flash->sector = (sector_t *)(flash->sector - sectors);
			continue;
		}
		
		if (th->function == T_StrobeFlash)
		{
			*save_p++ = tc_strobe;
			PADSAVEP();
			strobe = (strobe_t *)save_p;
			memcpy (strobe, th, sizeof(*strobe));
			save_p += sizeof(*strobe);
			strobe->sector = (sector_t *)(strobe->sector - sectors);
			continue;
		}
		
		if (th->function == T_Glow)
		{
			*save_p++ = tc_glow;
			PADSAVEP();
			glow = (glow_t *)save_p;
			memcpy (glow, th, sizeof(*glow));
			save_p += sizeof(*glow);
			glow->sector = (sector_t *)(glow->sector - sectors);
			continue;
		}
    }
	
    // add a terminating marker
    *save_p++ = tc_endspecials;	
}


//
// P_UnArchiveSpecials
//
void P_UnArchiveSpecials (void)
{
    byte			tclass;
    ceiling_t*		ceiling;
    vldoor_t*		door;
    floormove_t*	floor;
    plat_t*			plat;
    lightflash_t*	flash;
    strobe_t*		strobe;
    glow_t*			glow;
	
	
    // read in saved thinkers
    while (1)
    {
		tclass = *save_p++;
		switch (tclass)
		{
		case tc_endspecials:
			return;	// end of list
			
		case tc_ceiling:
			PADSAVEP();
			ceiling = Z_Malloc (sizeof(*ceiling), PU_LEVEL, NULL);
			memcpy (ceiling, save_p, sizeof(*ceiling));
			save_p += sizeof(*ceiling);
			ceiling->sector = &sectors[(int)ceiling->sector];
			ceiling->sector->specialdata = ceiling;
			
			if (ceiling->thinker.function)
				ceiling->thinker.function = T_MoveCeiling;
			
			P_AddThinker (&ceiling->thinker);
			P_AddActiveCeiling(ceiling);
			break;
			
		case tc_door:
			PADSAVEP();
			door = Z_Malloc (sizeof(*door), PU_LEVEL, NULL);
			memcpy (door, save_p, sizeof(*door));
			save_p += sizeof(*door);
			door->sector = &sectors[(int)door->sector];
			door->sector->specialdata = door;
			door->thinker.function = T_VerticalDoor;
			P_AddThinker (&door->thinker);
			break;
			
		case tc_floor:
			PADSAVEP();
			floor = Z_Malloc (sizeof(*floor), PU_LEVEL, NULL);
			memcpy (floor, save_p, sizeof(*floor));
			save_p += sizeof(*floor);
			floor->sector = &sectors[(int)floor->sector];
			floor->sector->specialdata = floor;
			floor->thinker.function = T_MoveFloor;
			P_AddThinker (&floor->thinker);
			break;
			
		case tc_plat:
			PADSAVEP();
			plat = Z_Malloc (sizeof(*plat), PU_LEVEL, NULL);
			memcpy (plat, save_p, sizeof(*plat));
			save_p += sizeof(*plat);
			plat->sector = &sectors[(int)plat->sector];
			plat->sector->specialdata = plat;
			
			if (plat->thinker.function)
				plat->thinker.function = T_PlatRaise;
			
			P_AddThinker (&plat->thinker);
			P_AddActivePlat(plat);
			break;
			
		case tc_flash:
			PADSAVEP();
			flash = Z_Malloc (sizeof(*flash), PU_LEVEL, NULL);
			memcpy (flash, save_p, sizeof(*flash));
			save_p += sizeof(*flash);
			flash->sector = &sectors[(int)flash->sector];
			flash->thinker.function = T_LightFlash;
			P_AddThinker (&flash->thinker);
			break;
			
		case tc_strobe:
			PADSAVEP();
			strobe = Z_Malloc (sizeof(*strobe), PU_LEVEL, NULL);
			memcpy (strobe, save_p, sizeof(*strobe));
			save_p += sizeof(*strobe);
			strobe->sector = &sectors[(int)strobe->sector];
			strobe->thinker.function = T_StrobeFlash;
			P_AddThinker (&strobe->thinker);
			break;
			
		case tc_glow:
			PADSAVEP();
			glow = Z_Malloc (sizeof(*glow), PU_LEVEL, NULL);
			memcpy (glow, save_p, sizeof(*glow));
			save_p += sizeof(*glow);
			glow->sector = &sectors[(int)glow->sector];
			glow->thinker.function = T_Glow;
			P_AddThinker (&glow->thinker);
			break;
			
		default:
			I_Error ("P_UnarchiveSpecials:Unknown tclass %i "
				"in savegame",tclass);
		}
    }
}

