// Emacs style mode select   -*- C++ -*- 
//-----------------------------------------------------------------------------
//
// $Id:$
//
// Copyright (C) 1993-1996 by id Software, Inc.
//
// This source is available for distribution and/or modification
// only under the terms of the DOOM Source Code License as
// published by id Software. All rights reserved.
//
// The source is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// FITNESS FOR A PARTICULAR PURPOSE. See the DOOM Source Code License
// for more details.
//
// $Log:$
//
// DESCRIPTION:
//	Player related stuff.
//	Bobbing POV/weapon, movement.
//	Pending weapon.
//
//-----------------------------------------------------------------------------


static const char
rcsid[] = "$Id: p_user.c,v 1.3 1997/01/28 22:08:29 b1 Exp $";

#include <math.h>

#include "doomdef.h"
#include "d_config.h"
#include "d_event.h"

#include "p_local.h"

#include "doomstat.h"



// Index of the special effects (INVUL inverse) map.
#define INVERSECOLORMAP		32


//
// Movement.
//

// 16 pixels of bob
#define MAXBOB	0x100000	

boolean		onground;


//
// P_Thrust
// Moves the given origin along a given angle.
//
void
P_Thrust
( player_t*	player,
  angle_t	angle,
  fixed_t	move ) 
{
    angle >>= ANGLETOFINESHIFT;
    
    player->plr->mo->momx += FixedMul(move,finecosine[angle]); 
    player->plr->mo->momy += FixedMul(move,finesine[angle]);
}




//
// P_CalcHeight
// Calculate the walking / running height adjustment
//
void P_CalcHeight (player_t* player) 
{
    int		angle;
    fixed_t	bob;
    
    // Regular movement bobbing
    // (needs to be calculated for gun swing
    // even if not on ground)
    // OPTIMIZE: tablify angle
    // Note: a LUT allows for effects
    //  like a ramp with low health.
    player->bob =
		FixedMul (player->plr->mo->momx, player->plr->mo->momx)
		+ FixedMul (player->plr->mo->momy,player->plr->mo->momy);
    
    player->bob >>= 2;
	
    if (player->bob>MAXBOB)
		player->bob = MAXBOB;
	
    if ((player->cheats & CF_NOMOMENTUM) || !onground)
    {
		player->plr->viewz = player->plr->mo->z + VIEWHEIGHT;
		
		if (player->plr->viewz > player->plr->mo->ceilingz-4*FRACUNIT)
			player->plr->viewz = player->plr->mo->ceilingz-4*FRACUNIT;
		
		player->plr->viewz = player->plr->mo->z + player->viewheight;
		return;
    }
	
    angle = (FINEANGLES/20*leveltime)&FINEMASK;
    bob = FixedMul ( player->bob/2, finesine[angle]);
	
    
    // move viewheight
    if (player->playerstate == PST_LIVE)
    {
		player->viewheight += player->deltaviewheight;
		
		if (player->viewheight > VIEWHEIGHT)
		{
			player->viewheight = VIEWHEIGHT;
			player->deltaviewheight = 0;
		}
		
		if (player->viewheight < VIEWHEIGHT/2)
		{
			player->viewheight = VIEWHEIGHT/2;
			if (player->deltaviewheight <= 0)
				player->deltaviewheight = 1;
		}
		
		if (player->deltaviewheight)	
		{
			player->deltaviewheight += FRACUNIT/4;
			if (!player->deltaviewheight)
				player->deltaviewheight = 1;
		}
    }
    player->plr->viewz = player->plr->mo->z + player->viewheight + bob;
	
    if (player->plr->viewz > player->plr->mo->ceilingz-4*FRACUNIT)
		player->plr->viewz = player->plr->mo->ceilingz-4*FRACUNIT;
}



//
// P_MovePlayer
//
void P_MovePlayer (player_t* player)
{
    ticcmd_t*		cmd;
	
    cmd = &player->cmd;
	
    player->plr->mo->angle += (cmd->angleturn<<16);

    // Do not let the player control movement
    //  if not onground.
    onground = (player->plr->mo->z <= player->plr->mo->floorz);
	if(player->plr->mo->onmobj && !onground)
	{
		mobj_t *on = player->plr->mo->onmobj;
		onground = (player->plr->mo->z <= on->z + on->height);
	}
	
    if (cmd->forwardmove && onground)
		P_Thrust (player, player->plr->mo->angle, cmd->forwardmove*2048);
    
    if (cmd->sidemove && onground)
		P_Thrust (player, player->plr->mo->angle-ANG90, cmd->sidemove*2048);

    if ( (cmd->forwardmove || cmd->sidemove) 
		&& player->plr->mo->state == &states[S_PLAY] )
    {
		P_SetMobjState (player->plr->mo, S_PLAY_RUN1);
    }

	if(cfg.jumpEnabled 
		&& onground
		&& !(cmd->buttons & BT_SPECIAL)
		&& !(cmd->buttons & BT_CHANGE)
		&& cmd->buttons & BT_JUMP
		&& !player->jumptics
		&& !(demoplayback || demorecording))
	{
		// Jump, then!
		player->plr->mo->momz = FRACUNIT * (netgame? 9 : cfg.jumpPower);
		player->jumptics = 24;
	}

	// Lookcenter flag?
	if(cmd->buttons & BT_LOOKCENTER) player->centering = true;
	// Forced looking?
	if(player->centering)
	{
		if(player->plr->lookdir > 0)
		{
			player->plr->lookdir -= 8;			
		}
		else if(player->plr->lookdir < 0)
		{
			player->plr->lookdir += 8;
		}
		if(fabs(player->plr->lookdir) <= 8)
		{
			player->plr->lookdir = 0;
			player->centering = false;
			// View centered.
		}
	}
	// Look up/down using the delta.
	else if(cmd->lookdirdelta)
	{
		float fd = cmd->lookdirdelta / DELTAMUL;
		float delta = fd * fd;	// fd^2
		if(cmd->lookdirdelta < 0) delta = -delta;
		player->plr->lookdir += delta;
	}
	
	// 110 corresponds 85 degrees. Go figure...
	if(player->plr->lookdir > 110) player->plr->lookdir = 110;
	if(player->plr->lookdir < -110) player->plr->lookdir = -110;
}	



//
// P_DeathThink
// Fall on your face when dying.
// Decrease POV height to floor height.
//
#define ANG5   	(ANG90/18)

void P_DeathThink (player_t* player)
{
    angle_t		angle;
    angle_t		delta;

    P_MovePsprites (player);
	
    // fall to the ground
    if (player->viewheight > 6*FRACUNIT)
	player->viewheight -= FRACUNIT;

    if (player->viewheight < 6*FRACUNIT)
	player->viewheight = 6*FRACUNIT;

    player->deltaviewheight = 0;
    onground = (player->plr->mo->z <= player->plr->mo->floorz);
    P_CalcHeight (player);
	
    if (player->attacker && player->attacker != player->plr->mo)
    {
	angle = R_PointToAngle2 (player->plr->mo->x,
				 player->plr->mo->y,
				 player->attacker->x,
				 player->attacker->y);
	
	delta = angle - player->plr->mo->angle;
	
	if (delta < ANG5 || delta > (unsigned)-ANG5)
	{
	    // Looking at killer,
	    //  so fade damage flash down.
	    player->plr->mo->angle = angle;

	    if (player->damagecount)
		player->damagecount--;
	}
	else if (delta < ANG180)
	    player->plr->mo->angle += ANG5;
	else
	    player->plr->mo->angle -= ANG5;
    }
    else if (player->damagecount)
	player->damagecount--;
	

    if (player->cmd.buttons & BT_USE)
	player->playerstate = PST_REBORN;
}

#define LIST_BEGIN		-1
#define LIST_END		-2

weapontype_t P_PlayerFindWeapon(player_t *player, boolean next)
{
/*		if (newweapon == wp_fist
			&& player->weaponowned[wp_chainsaw]
			&& !(player->readyweapon == wp_chainsaw
			&& player->powers[pw_strength]))
		{
			newweapon = wp_chainsaw;
		}
		
		if ( (gamemode == commercial)
			&& newweapon == wp_shotgun 
			&& player->weaponowned[wp_supershotgun]
			&& player->readyweapon != wp_supershotgun)
		{
			newweapon = wp_supershotgun;
		}

		if (player->weaponowned[newweapon]
			&& newweapon != player->readyweapon)
		{
			// Do not go to plasma or BFG in shareware,
			//  even if cheated.
			if ((newweapon != wp_plasma
				&& newweapon != wp_bfg)
				|| (gamemode != shareware) )
			{
				player->pendingweapon = newweapon;
			}
		}
*/	
	int sw_list[] = 
	{ 
		wp_fist, wp_pistol, wp_shotgun, wp_chaingun, wp_missile, wp_chainsaw
	};
	int reg_list[] = 
	{ 
		wp_fist, wp_pistol, wp_shotgun, wp_chaingun, wp_missile, wp_plasma, 
		wp_bfg, wp_chainsaw
	};
	int dm2_list[] =
	{
		wp_fist, wp_pistol, wp_shotgun, wp_supershotgun, wp_chaingun, 
		wp_missile,	wp_plasma, wp_bfg, wp_chainsaw
	};
	int *list, num, i;

	if(gamemode == shareware)
	{
		list = sw_list;
		num = 6;
	}
	else if(gamemode == commercial)
	{
		list = dm2_list;
		num = 9;
	}
	else
	{
		list = reg_list;
		num = 8;
	}
	// Find the current position in the weapon list.
	for(i=0; i<num; i++) if(list[i] == player->readyweapon) break;
	// Locate the next or previous weapon owned by the player.
	for(;;)
	{
		// Move the iterator.
		if(next) i++; else i--;
		if(i < 0) i = num-1;
		else if(i > num-1) i = 0;
		// Have we circled around?
		if(list[i] == player->readyweapon) break;
		// A valid weapon?
		if(player->weaponowned[list[i]]) break;
	}
	return list[i];
}

//
// P_PlayerThink
//
void P_PlayerThink (player_t* player)
{
    ticcmd_t*		cmd;
    weapontype_t	newweapon;
	
    // fixme: do this in the cheat code
    if (player->cheats & CF_NOCLIP)
		player->plr->mo->flags |= MF_NOCLIP;
    else
		player->plr->mo->flags &= ~MF_NOCLIP;
    
    // chain saw run forward
    cmd = &player->cmd;
    if (player->plr->mo->flags & MF_JUSTATTACKED)
    {
		cmd->angleturn = 0;
		cmd->forwardmove = 0xc800/512;
		cmd->sidemove = 0;
		player->plr->mo->flags &= ~MF_JUSTATTACKED;
    }
	
	
    if (player->playerstate == PST_DEAD)
    {
		P_DeathThink (player);
		return;
    }
    
    // Move around.
    // Reactiontime is used to prevent movement
    //  for a bit after a teleport.
    if (player->plr->mo->reactiontime)
		player->plr->mo->reactiontime--;
    else
		P_MovePlayer (player);
    
    P_CalcHeight (player);
	
    if (player->plr->mo->subsector->sector->special)
		P_PlayerInSpecialSector (player);
    
	if(player->jumptics) player->jumptics--;

	if(cmd->buttons & BT_SPECIAL)
	{
		// There might be a special weapon change.
		if(cmd->buttons & (BTS_NEXTWEAPON | BTS_PREVWEAPON) &&
			!(cmd->buttons & (BTS_SAVEGAME|BTS_PAUSE)))
		{
			player->pendingweapon = P_PlayerFindWeapon(player, 
				(cmd->buttons & BTS_NEXTWEAPON) != 0);
		}
		cmd->buttons = 0;
	}

    // Check for weapon change.
    if (cmd->buttons & BT_CHANGE)
    {
		// The actual changing of the weapon is done
		//  when the weapon psprite can do it
		//  (read: not in the middle of an attack).
		newweapon = (cmd->buttons&BT_WEAPONMASK)>>BT_WEAPONSHIFT;

		if (newweapon == wp_fist
			&& player->weaponowned[wp_chainsaw]
			&& !(player->readyweapon == wp_chainsaw
			&& player->powers[pw_strength]))
		{
			newweapon = wp_chainsaw;
		}
		
		if ( (gamemode == commercial)
			&& newweapon == wp_shotgun 
			&& player->weaponowned[wp_supershotgun]
			&& player->readyweapon != wp_supershotgun)
		{
			newweapon = wp_supershotgun;
		}

		if (player->weaponowned[newweapon]
			&& newweapon != player->readyweapon)
		{
			// Do not go to plasma or BFG in shareware,
			//  even if cheated.
			if ((newweapon != wp_plasma
				&& newweapon != wp_bfg)
				|| (gamemode != shareware) )
			{
				player->pendingweapon = newweapon;
			}
		}
    }
    
    // check for use
    if (cmd->buttons & BT_USE)
    {
		if (!player->usedown)
		{
			P_UseLines (player);
			player->usedown = true;
		}
    }
    else
		player->usedown = false;
    
    // cycle psprites
    P_MovePsprites (player);
    
    // Counters, time dependend power ups.
	
    // Strength counts up to diminish fade.
    if (player->powers[pw_strength])
		player->powers[pw_strength]++;	
	
    if (player->powers[pw_invulnerability])
		player->powers[pw_invulnerability]--;
	
    if (player->powers[pw_invisibility])
		if (! --player->powers[pw_invisibility] )
			player->plr->mo->flags &= ~MF_SHADOW;
		
		if (player->powers[pw_infrared])
			player->powers[pw_infrared]--;
		
		if (player->powers[pw_ironfeet])
			player->powers[pw_ironfeet]--;
		
		if (player->damagecount)
			player->damagecount--;
		
		if (player->bonuscount)
			player->bonuscount--;
		
		
		// Handling colormaps.
/*		if (player->powers[pw_invulnerability])
		{
			if (player->powers[pw_invulnerability] > 4*32
				|| (player->powers[pw_invulnerability]&8) )
				player->plr->fixedcolormap = INVERSECOLORMAP;
			else
				player->plr->fixedcolormap = 0;
		}
		else if (player->powers[pw_infrared])	
		{
			if (player->powers[pw_infrared] > 4*32
				|| (player->powers[pw_infrared]&8) )
			{
				// almost full bright
				player->plr->fixedcolormap = 1;
			}
			else
				player->plr->fixedcolormap = 0;
		}
		else
			player->plr->fixedcolormap = 0;*/
}


