#include "DoomDef.h"
#include "P_local.h"
#include "H_Net.h"
#include "soundst.h"
#include "settings.h"

// TYPES --------------------------------------------------------------------

// servdata_t can be *max* 12 bytes long (3 ints).
typedef struct
{
	unsigned char	nomonsters : 1;
	unsigned char	respawn : 1;
	unsigned char	skill : 4;
	unsigned char	map : 5;
	unsigned char	episode : 3;
	unsigned char	deathmatch : 1;
	unsigned char	slot : 3;
	unsigned char	jumping : 1;
} servdata_t;

// PUBLIC DATA --------------------------------------------------------------

/*boolean		netDeathmatch = false, netNomonsters = false, netRespawn = false;
boolean		netJumping = true;
byte		netEpisode = 1, netMap = 1, netSkill = sk_medium, netSlot = 0;
byte		netColor = 0;
int			PlayerColor[MAXPLAYERS];*/

// CODE ---------------------------------------------------------------------

void UpdateServerData(int set)
{
	servdata_t svd;

	if(set)
	{
		// Pack the data into the server data word.
		svd.deathmatch = cfg.netDeathmatch;
		svd.map = cfg.netMap;
		svd.episode = cfg.netEpisode;
		svd.nomonsters = cfg.netNomonsters;
		svd.respawn = cfg.netRespawn;
		svd.skill = cfg.netSkill;
		svd.slot = cfg.netSlot;
		svd.jumping = cfg.netJumping;
		gi.NetSetServerData(&svd, sizeof(svd));
	}
	else
	{
		gi.NetGetServerData(&svd, sizeof(svd));
		// Unpack the data.
		cfg.netDeathmatch = svd.deathmatch;
		cfg.netMap = svd.map;
		cfg.netEpisode = svd.episode;
		cfg.netNomonsters = svd.nomonsters;
		cfg.netRespawn = svd.respawn;
		cfg.netSkill = svd.skill;
		cfg.netSlot = svd.slot;
		cfg.netJumping = svd.jumping;
	}
}

int H_NetServerOpen(int before)
{
	if(before) 
		UpdateServerData(true);
	else
	{
		plrdata_t pd = { cfg.netColor };
		gi.NetSetPlayerData(&pd, sizeof(pd));
	}
	return true;
}

int H_NetServerClose(int before)
{
	if(!before)
	{
		// Restore normal game state.
		deathmatch = false;
		nomonsters = false;

		P_SetMessage(&players[consoleplayer], "NETGAME ENDS", true);
		S_StartSound(NULL, sfx_dorcls);
	}
	return true;
}

int H_NetServerStarted(int before)
{
	int			i, server = gi.Get(DD_SERVER);
	plrdata_t	pd;

	if(before)
	{
		// If we are the server, we must update the game data.
		if(server) UpdateServerData(true);
		
		// Update our player data, for the last time.
		pd.color = cfg.netColor;
		gi.NetSetPlayerData(&pd, sizeof(pd));
		return true;
	}
	// For clients: read the final server data.
	if(!server)	UpdateServerData(false);

	// Let's read the data of each player.
	for(i=0; i<MAXPLAYERS; i++)
	{
		if(players[i].plr->ingame)
		{
			gi.NetGetPlayerData(i, &pd, sizeof(pd));
/*			PlayerClass[i] = pd.class;*/
			cfg.PlayerColor[i] = pd.color;
		}
	}

	// Set the game parameters.
	deathmatch = cfg.netDeathmatch;
	nomonsters = cfg.netNomonsters;
	respawnparm = cfg.netRespawn;
	cfg.jumpEnabled = cfg.netJumping;

	G_CheckDemoStatus();
	advancedemo = false;

	// Start the game.	
	if(!cfg.netSlot)
	{
//		G_StartNewInit();
		G_InitNew(cfg.netSkill, cfg.netEpisode, cfg.netMap);
	}
	else
	{
		char filename[16];
		sprintf(filename, SAVEGAMENAME"%c.hsg", cfg.netSlot-1);
		// Load the game from the slot.
		G_LoadGame(filename);
	}
	// Close the menu, the game begins!!
	MN_DeactivateMenu();
	return true;
}

int H_NetConnect(int before)
{
	plrdata_t	pd = { cfg.netColor };

	// We do nothing before the actual connection is made.
	if(before) return true;
	
	// After connecting we get the game data from the server.
	UpdateServerData(false);

	// Set the player data.
	gi.NetSetPlayerData(&pd, sizeof(pd));

	// Some diagnostics.
	gi.Message("Netgame: map%i, episode%i, skill%i, dm%i, rs%i, !mnst%i, slot%i, jump%i\n",
		cfg.netMap, cfg.netEpisode, cfg.netSkill, cfg.netDeathmatch, 
		cfg.netRespawn, cfg.netNomonsters, cfg.netSlot, cfg.netJumping);
	return true;
}

int H_NetDisconnect(int before)
{
	if(before) return true;

	// Restore normal game state.
	deathmatch = false;
	nomonsters = false;
	return true;
}

int H_NetPlayerEvent(int plrNumber, int peType, void *data)
{
	char	msgbuff[256];

	// If this isn't a netgame, we won't react.
	if(!gi.Get(DD_NETGAME)) return true;

	if(peType == DDPE_EXIT)
	{
		// A player has exited a netgame?
		strcpy(msgbuff, "PLAYER 1 LEFT THE GAME");
		msgbuff[7] += plrNumber;
		P_SetMessage(&players[consoleplayer], msgbuff, true);
		S_StartSound(NULL, sfx_chat);
	}
	// Don't show local chat messages.
	else if(peType == DDPE_CHAT_MESSAGE && plrNumber != consoleplayer)
	{
		int i, num, oldecho = cfg.echoMsg;
		// Count the number of players.
		for(i=num=0; i<MAXPLAYERS; i++)
			if(players[i].plr->ingame) num++;
		// If there are more than two players, include the name of
		// the player who sent this.
		if(num > 2)
			sprintf(msgbuff, "%s: %s", gi.GetPlayerName(plrNumber), data);
		else
			strcpy(msgbuff, data);
		MN_TextFilter(msgbuff);
		// The chat message is already echoed by the console.
		cfg.echoMsg = false;
		P_SetMessage(&players[consoleplayer], msgbuff, true);
		cfg.echoMsg = oldecho;
		S_StartSound(NULL, sfx_chat);
	}
	return true;
}
